#ifndef BASEMOD_H
#define BASEMOD_H

#define PB_ERROR 1
#define PB_OK 0
#define PB_EMPTY 1
#define PB_SOLVED 2

typedef struct{
	/*pointer to an array containing the sudoku. the variable numbers are at the positions pos[i]*/
	int* sudoku;
	/*the width of the sudoku, a square*/
	int size;
	/*array containing the indices of the variable numbers*/
	int* pos;
	/*the number of variable numbers*/
	int free;
	/*arrays containing the upper and lower bounds for this subtree to check for the ith variable number, it is always safe to assume lower[i] <= sudoku[pos[i]] < upper[i]*/
	int* upper;
	int* lower;
	/* indicates the current working level*/
	int clevel;
}subtree_t;

/*st is a pointer to an uninitialized subtree struct, sudoku is an array of size*size length containing the fixed numbers and zeros. if subtree init returns PB_OK, st is an initialized subtree, otherwise not*/
int subtree_init(subtree_t *st, int size, int *sudoku);

/*st is a pointer to an initialized subtree struct. if subtree_destroy returns PB_OK, st is an uninitialized subtree*/
int subtree_destroy(subtree_t *st);

int subtree_work(subtree_t *st);

int subtree_print(subtree_t *st);

int subtree_check(subtree_t *st);

int subtree_copy(subtree_t *stin, subtree_t *stout);

int subtree_split(subtree_t *stin, subtree_t *stout);
#endif
