#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>
#include "common/basemod.h"

int subtree_sendrecv(subtree_t *stin, subtree_t *stout, int dest){
	int i=0;
	MPI_Request request[7];

	MPI_Isend(&stout->size,1,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);
	MPI_Isend(&stout->free,1,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);
	MPI_Isend(&stout->clevel,1,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);
	MPI_Isend(stout->sudoku,stout->size*stout->size,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);
	MPI_Isend(stout->pos,stout->free,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);
	MPI_Isend(stout->lower,stout->free,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);
	MPI_Isend(stout->upper,stout->free,MPI_INT,dest,1,MPI_COMM_WORLD,&request[i++]);

	MPI_Recv(&stin->size,1,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(&stin->free,1,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(&stin->clevel,1,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->sudoku,stin->size*stin->size,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->pos,stin->free,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->lower,stin->free,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->upper,stin->free,MPI_INT,MPI_ANY_SOURCE,1,MPI_COMM_WORLD,NULL);
	MPI_Waitall(7,request,NULL);
	return(0);
}

int subtree_send(subtree_t *stout,int dest){
	MPI_Send(&stout->size,1,MPI_INT,dest,1,MPI_COMM_WORLD);
	MPI_Send(&stout->free,1,MPI_INT,dest,1,MPI_COMM_WORLD);
	MPI_Send(&stout->clevel,1,MPI_INT,dest,1,MPI_COMM_WORLD);
	MPI_Send(stout->sudoku,stout->size*stout->size,MPI_INT,dest,1,MPI_COMM_WORLD);
	MPI_Send(stout->pos,stout->free,MPI_INT,dest,1,MPI_COMM_WORLD);
	MPI_Send(stout->lower,stout->free,MPI_INT,dest,1,MPI_COMM_WORLD);
	MPI_Send(stout->upper,stout->free,MPI_INT,dest,1,MPI_COMM_WORLD);
	return(0);
}

int subtree_recv(subtree_t *stin,int source){
	MPI_Recv(&stin->size,1,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(&stin->free,1,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(&stin->clevel,1,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->sudoku,stin->size*stin->size,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->pos,stin->free,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->lower,stin->free,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	MPI_Recv(stin->upper,stin->free,MPI_INT,source,1,MPI_COMM_WORLD,NULL);
	return(0);
}

int randomperm(int rank, int size){
	int j,probdest,b_flag,Delta,lpermsize,next,prev,permsize;
	int localperm[size];
	MPI_Request requests[size];
	MPI_Status status;


	prev = (rank == 0) ? MPI_PROC_NULL : rank-1;
	next = (rank == size-1) ? MPI_PROC_NULL : rank+1;

	probdest = rand()%size;

	/*tag 17 is reserved for communication in connection with random permutation*/
	MPI_Isend(&rank,1,MPI_INT,probdest,17,MPI_COMM_WORLD,&requests[0]);
	MPI_Barrier(MPI_COMM_WORLD);

	lpermsize=0;
	MPI_Iprobe(MPI_ANY_SOURCE,17,MPI_COMM_WORLD,&b_flag,&status);
	while(b_flag){
		MPI_Recv(&localperm[lpermsize],1,MPI_INT,status.MPI_SOURCE,17,MPI_COMM_WORLD,NULL);
		lpermsize++;
		MPI_Iprobe(MPI_ANY_SOURCE,17,MPI_COMM_WORLD,&b_flag,&status);
	}
	MPI_Allreduce(&lpermsize,&permsize,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
	while(permsize!=size){
		MPI_Iprobe(MPI_ANY_SOURCE,17,MPI_COMM_WORLD,&b_flag,&status);
		if(b_flag){
			MPI_Recv(&localperm[lpermsize],1,MPI_INT,status.MPI_SOURCE,17,MPI_COMM_WORLD,NULL);
			lpermsize++;
		}
		MPI_Allreduce(&lpermsize,&permsize,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
	}
	MPI_Barrier(MPI_COMM_WORLD);
	

	Delta = 0;
	MPI_Recv(&Delta,1,MPI_INT,prev,18,MPI_COMM_WORLD,NULL);
	Delta += lpermsize;
	MPI_Send(&Delta,1,MPI_INT,next,18,MPI_COMM_WORLD);
	Delta -= lpermsize;
	MPI_Barrier(MPI_COMM_WORLD);

	for(j=0;j<lpermsize;j++){
		MPI_Isend(&localperm[j],1,MPI_INT,Delta+j,18,MPI_COMM_WORLD,&requests[j]);
	}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Recv(&probdest,1,MPI_INT,MPI_ANY_SOURCE,18,MPI_COMM_WORLD,NULL);
	MPI_Waitall(lpermsize,requests,NULL);
	return(probdest);
}

int main(int argc, char** argv){
	int rank,size,sudsize,status,solved,empty,globsolved,globempty;
	int cycle,i,j,dim,dest,NoP,nstatus;
	int *sudoku,*nps,*periods;
	double Start,End,Calctime;
	subtree_t local,temp;
	MPI_Comm hypercube;
	MPI_Status dummystatus;

	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD,&size);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);
	srand(rank);

	NoP=atoi(argv[1]);

	if(rank==0)	Start = MPI_Wtime();

	dim =floor(log(size)/log(2));
	if(1<<dim != size){
		printf("Anzahl Prozesse muss eine Potenz von zwei sein!\n");
		MPI_Finalize();
		return(0);
	}
	nps = (int*)malloc(sizeof(int)*dim);
	periods = (int*)malloc(sizeof(int)*dim);
	for(i=0;i<dim;i++){
		nps[i] = 2;
		periods[i] = 1;
	}
	MPI_Cart_create(MPI_COMM_WORLD,dim,nps,periods,0,&hypercube);

	if(rank == 0){
		/* read in sudoku*/
		scanf("%d\n",&sudsize);
		MPI_Bcast(&sudsize,1,MPI_INT,0,MPI_COMM_WORLD);
		sudoku = (int*)malloc(sizeof(int)*sudsize*sudsize);
		for(i=0;i<sudsize;i++){
			for(j=0;j<sudsize;j++){
				scanf("%d\t",&sudoku[sudsize*i+j]);
			}
		}
		status = PB_OK;
	} else{
		MPI_Bcast(&sudsize,1,MPI_INT,0,MPI_COMM_WORLD);
		sudoku = (int*)malloc(sizeof(int)*sudsize*sudsize);
		status = PB_EMPTY;
	}

	MPI_Bcast(sudoku,sudsize*sudsize,MPI_INT,0,MPI_COMM_WORLD);

	subtree_init(&local,sudsize,sudoku);
	subtree_init(&temp,sudsize,sudoku);

//	if(rank==1) subtree_print(&local);
	globsolved = globempty = 0;
	cycle = 0;
	while(1){
		/*termination condition*/
		solved = (status == PB_SOLVED);
		empty = (status == PB_EMPTY);
		MPI_Allreduce(&solved,&globsolved,1,MPI_INT,MPI_LOR,MPI_COMM_WORLD);
		MPI_Allreduce(&empty,&globempty,1,MPI_INT,MPI_LAND,MPI_COMM_WORLD);
		if(globempty || globsolved) break;

		/*random permutation if necessary*/
		if(cycle == dim){
			dest = randomperm(rank,size);
			subtree_copy(&local,&temp);
			subtree_sendrecv(&local,&temp,dest);
			cycle = 0;
		}

		/*start next phase*/
		for(i=0;i<NoP;i++){
			if(status == PB_OK) status = subtree_work(&local);
		}
		MPI_Barrier(MPI_COMM_WORLD);
		/*communication with neighbours*/
		MPI_Cart_shift(hypercube,cycle,1,&dest,&dest);
		MPI_Sendrecv(&status,1,MPI_INT,dest,23,&nstatus,1,MPI_INT,dest,23,MPI_COMM_WORLD,&dummystatus);
		if(status == PB_OK && nstatus == PB_OK){
			//do nothing
		} else if(status == PB_OK && nstatus == PB_EMPTY){
			//send
			subtree_split(&local,&temp);
			subtree_send(&temp,dest);
		} else if(status == PB_EMPTY && nstatus == PB_OK){
			//receive
			subtree_recv(&local,dest);
			status = PB_OK;
		} else if(status == PB_EMPTY && nstatus == PB_EMPTY){
			//do nothing
		}
		cycle++;
	}

//	if(solved){
//		printf("Sudoku solved by process %d\n",rank);
//		subtree_print(&local);
//	} else if(globempty && rank == 0){
//		printf("Sudoku is unsolvable\n");
//	}

	if(rank==0){
		End = MPI_Wtime();
		Calctime = End-Start;
		printf("%f\n",Calctime);
	}


	free(sudoku);
	free(nps);
	free(periods);
	subtree_destroy(&local);
	subtree_destroy(&temp);
	MPI_Finalize();
	return(0);
}
